# ------------------------------------------------------------------------------------
# Helper to use PCL from outside project
#
# target_link_libraries(my_fabulous_target ROS_PCL_XXX_LIBRARIES) where XXX is the 
# upper cased xxx from : 
# 
# - common
# - sample_consensus
# - octree
# - io
# - kdtree
# - search
# - features
# - filters
# - keypoints
# - surface
# - registration
# - segmentation
# - tracking
# - visualization
#
# ROS_PCL_INCLUDE_DIRS is filled with PCL and available 3rdparty headers
# ROS_PCL_LIBRARY_DIRS is filled with PCL components libraries install directory and
# 3rdparty libraries paths
# 
#                                   www.pointclouds.org
#------------------------------------------------------------------------------------

### ---[ some useful macros
macro(ros_pcl_report_not_found _reason)
  unset(ROS_PCL_FOUND)
  unset(ROS_PCL_LIBRARIES)
  unset(ROS_PCL_INCLUDE_DIRS)
  unset(ROS_PCL_LIBRARY_DIRS)
  unset(ROS_PCL_DEFINITONS)
  if(ROS_PCL_FIND_REQUIRED)
    message(FATAL_ERROR ${_reason})
  elseif(NOT ROS_PCL_FIND_QUIETLY)
    message(WARNING ${_reason})
  endif()
  return()
endmacro(ros_pcl_report_not_found)

macro(ros_pcl_message)
  if(NOT ROS_PCL_FIND_QUIETLY)
    message(${ARGN})
  endif(NOT ROS_PCL_FIND_QUIETLY)
endmacro(ros_pcl_message)

# Remove duplicate libraries	
macro(ros_pcl_remove_duplicate_libraries _unfiltered_libraries _filtered_libraries)
  set(${_filtered_libraries})
  set(_debug_libraries)
  set(_optimized_libraries)
  set(_other_libraries)
  set(_waiting_for_debug 0)
  set(_waiting_for_optimized 0)
  set(_library_position -1)
  foreach(library ${${_unfiltered_libraries}})
    if("${library}" STREQUAL "debug")
      set(_waiting_for_debug 1)
    elseif("${library}" STREQUAL "optimized")
      set(_waiting_for_optimized 1)
    elseif(_waiting_for_debug)
      list(FIND _debug_libraries "${library}" library_position)
      if(library_position EQUAL -1)
        list(APPEND ${_filtered_libraries} debug ${library})
        list(APPEND _debug_libraries ${library})
      endif()
      set(_waiting_for_debug 0)
    elseif(_waiting_for_optimized)
      list(FIND _optimized_libraries "${library}" library_position)
      if(library_position EQUAL -1)
        list(APPEND ${_filtered_libraries} optimized ${library})
        list(APPEND _optimized_libraries ${library})
      endif()
      set(_waiting_for_optimized 0)
    else("${library}" STREQUAL "debug")
      list(FIND _other_libraries "${library}" library_position)
      if(library_position EQUAL -1)
        list(APPEND ${_filtered_libraries} ${library})
        list(APPEND _other_libraries ${library})
      endif()
    endif("${library}" STREQUAL "debug")
  endforeach(library)
endmacro(ros_pcl_remove_duplicate_libraries)

### ---[ 3rd party libraries
macro(find_boost)
  if(ROS_PCL_ALL_IN_ONE_INSTALLER)
    set(BOOST_ROOT "${ROS_PCL_ROOT}/3rdParty/Boost")
  elseif(NOT BOOST_INCLUDEDIR)
    set(BOOST_INCLUDEDIR "/usr/include")
  endif(ROS_PCL_ALL_IN_ONE_INSTALLER)
  # use static Boost in Windows
  if(WIN32)
    set(Boost_USE_STATIC_LIBS ON)
    set(Boost_USE_STATIC ON)
  endif(WIN32)
  if(${CMAKE_VERSION} VERSION_LESS 2.8.5)
    SET(Boost_ADDITIONAL_VERSIONS "1.43" "1.43.0" "1.44" "1.44.0" "1.45" "1.45.0" "1.46.1" "1.46.0" "1.46" "1.47" "1.47.0")
  else(${CMAKE_VERSION} VERSION_LESS 2.8.5)
    SET(Boost_ADDITIONAL_VERSIONS "1.47" "1.47.0")
  endif(${CMAKE_VERSION} VERSION_LESS 2.8.5)
  # Disable the config mode of find_package(Boost)
  set(Boost_NO_BOOST_CMAKE ON)
  find_package(Boost 1.40.0 ${QUIET_} COMPONENTS system filesystem thread date_time iostreams)
  set(BOOST_FOUND ${Boost_FOUND})
  set(BOOST_INCLUDE_DIRS "${Boost_INCLUDE_DIR}")
  set(BOOST_LIBRARY_DIRS "${Boost_LIBRARY_DIRS}")
  set(BOOST_LIBRARIES ${Boost_LIBRARIES})
  if(WIN32)
    set(BOOST_DEFINITIONS ${BOOST_DEFINITIONS} -DBOOST_ALL_NO_LIB)
  endif(WIN32)
endmacro(find_boost)

#remove this as soon as eigen is shipped with FindEigen.cmake
macro(find_eigen)
  if(ROS_PCL_ALL_IN_ONE_INSTALLER)
    set(EIGEN_ROOT "${ROS_PCL_ROOT}/3rdParty/Eigen")
  elseif(NOT EIGEN_ROOT)
    get_filename_component(EIGEN_ROOT "/usr/include/eigen3" PATH)
  endif(ROS_PCL_ALL_IN_ONE_INSTALLER)
  if(PKG_CONFIG_FOUND)
    pkg_check_modules(PC_EIGEN eigen3)
  endif(PKG_CONFIG_FOUND)
  find_path(EIGEN_INCLUDE_DIRS Eigen/Core
    HINTS ${PC_EIGEN_INCLUDEDIR} ${PC_EIGEN_INCLUDE_DIRS} 
          "${EIGEN_ROOT}" "$ENV{EIGEN_ROOT}"
    PATHS "$ENV{PROGRAMFILES}/Eigen 3.0.0" "$ENV{PROGRAMW6432}/Eigen 3.0.0"
          "$ENV{PROGRAMFILES}/Eigen" "$ENV{PROGRAMW6432}/Eigen"   
    PATH_SUFFIXES eigen3 include/eigen3 include)
  find_package_handle_standard_args(eigen DEFAULT_MSG EIGEN_INCLUDE_DIRS)
  set(EIGEN_DEFINITIONS ${EIGEN_DEFINITIONS} -DEIGEN_USE_NEW_STDVECTOR 
      -DEIGEN_YES_I_KNOW_SPARSE_MODULE_IS_NOT_STABLE_YET)
endmacro(find_eigen)

#remove this as soon as qhull is shipped with FindQhull.cmake
macro(find_qhull)
  if(ROS_PCL_ALL_IN_ONE_INSTALLER)
    set(QHULL_ROOT "${ROS_PCL_ROOT}/3rdParty/Qhull")
  elseif(NOT QHULL_ROOT)
    get_filename_component(QHULL_ROOT "/usr/include/qhull" PATH)
  endif(ROS_PCL_ALL_IN_ONE_INSTALLER)
  set(QHULL_MAJOR_VERSION 6)

  find_path(QHULL_INCLUDE_DIRS
    NAMES libqhull/libqhull.h qhull.h
    HINTS "${QHULL_ROOT}" "$ENV{QHULL_ROOT}"
    PATHS "$ENV{PROGRAMFILES}/qhull 6.2.0.1373" "$ENV{PROGRAMW6432}/qhull 6.2.0.1373" 
          "$ENV{PROGRAMFILES}/qhull" "$ENV{PROGRAMW6432}/qhull" 
    PATH_SUFFIXES qhull src/libqhull libqhull include)

  # Most likely we are on windows so prefer static libraries over shared ones (Mourad's recommend)
  find_library(QHULL_LIBRARY 
    NAMES qhullstatic qhull qhull${QHULL_MAJOR_VERSION}
    HINTS "${QHULL_ROOT}" "$ENV{QHULL_ROOT}"
    PATHS "$ENV{PROGRAMFILES}/qhull 6.2.0.1373" "$ENV{PROGRAMW6432}/qhull 6.2.0.1373" 
          "$ENV{PROGRAMFILES}/qhull" "$ENV{PROGRAMW6432}/qhull" 
    PATH_SUFFIXES project build bin lib)
  
  find_library(QHULL_LIBRARY_DEBUG 
    NAMES qhullstatic_d qhull_d qhull${QHULL_MAJOR_VERSION}_d 
          qhull qhull${QHULL_MAJOR_VERSION}
    HINTS "${QHULL_ROOT}" "$ENV{QHULL_ROOT}"
    PATHS "$ENV{PROGRAMFILES}/qhull 6.2.0.1373" "$ENV{PROGRAMW6432}/qhull 6.2.0.1373" 
          "$ENV{PROGRAMFILES}/qhull" "$ENV{PROGRAMW6432}/qhull" 
    PATH_SUFFIXES project build bin lib)
  
  find_package_handle_standard_args(qhull DEFAULT_MSG QHULL_LIBRARY QHULL_INCLUDE_DIRS)

  if(QHULL_FOUND)
    get_filename_component(QHULL_LIBRARY_PATH ${QHULL_LIBRARY} PATH)
    if(QHULL_LIBRARY_DEBUG)
      set(QHULL_LIBRARIES optimized ${QHULL_LIBRARY} debug ${QHULL_LIBRARY_DEBUG})
      get_filename_component(QHULL_LIBRARY_DEBUG_PATH ${QHULL_LIBRARY_DEBUG} PATH)
      set(QHULL_LIBRARY_DIRS ${QHULL_LIBRARY_PATH} ${QHULL_LIBRARY_DEBUG_PATH}) 
    else(QHULL_LIBRARY_DEBUG)
      set(QHULL_LIBRARIES ${QHULL_LIBRARY})
      set(QHULL_LIBRARY_DIRS ${QHULL_LIBRARY_PATH}) 
    endif(QHULL_LIBRARY_DEBUG)
  endif(QHULL_FOUND)
endmacro(find_qhull)

#remove this as soon as openni-dev is shipped with FindOpenni.cmake
macro(find_openni)
  if(NOT OPENNI_ROOT AND ("" STREQUAL "ON"))
    set(OPENNI_INCLUDE_DIRS_HINT "OPENNI_INCLUDE_DIR-NOTFOUND")
    set(OPENNI_LIBRARY_HINT "/usr/lib64/libOpenNI.so")
  endif(NOT OPENNI_ROOT AND ("" STREQUAL "ON"))

  if(PKG_CONFIG_FOUND)
    pkg_check_modules(PC_OPENNI openni-dev)
  endif(PKG_CONFIG_FOUND)
  find_path(OPENNI_INCLUDE_DIRS XnStatus.h
    HINTS ${PC_OPENNI_INCLUDEDIR} ${PC_OPENNI_INCLUDE_DIRS} 
          "${OPENNI_ROOT}" "$ENV{OPENNI_ROOT}"
    PATHS "$ENV{PROGRAMFILES}/OpenNI/Include" "$ENV{PROGRAMW6432}/OpenNI/Include" "${OPENNI_INCLUDE_DIRS_HINT}"
    PATH_SUFFIXES include/openni Include)
  #add a hint so that it can find it without the pkg-config
  find_library(OPENNI_LIBRARY 
    NAMES OpenNI64 OpenNI 
    HINTS ${PC_OPENNI_LIBDIR} ${PC_OPENNI_LIBRARY_DIRS} 
          "${OPENNI_ROOT}" "$ENV{OPENNI_ROOT}"
    PATHS "$ENV{PROGRAMFILES}/OpenNI/Lib" "$ENV{PROGRAMW6432}/OpenNI/Lib64" "${OPENNI_LIBRARY_HINT}"
    PATH_SUFFIXES lib Lib Lib64)

  find_package_handle_standard_args(openni DEFAULT_MSG OPENNI_LIBRARY OPENNI_INCLUDE_DIRS)

  if(OPENNI_FOUND)
    get_filename_component(OPENNI_LIBRARY_PATH ${OPENNI_LIBRARY} PATH)
    set(OPENNI_LIBRARY_DIRS ${OPENNI_LIBRARY_PATH}) 
    set(OPENNI_LIBRARIES "${OPENNI_LIBRARY}")
  endif(OPENNI_FOUND)
endmacro(find_openni)

#remove this as soon as flann is shipped with FindFlann.cmake
macro(find_flann)
  if(ROS_PCL_ALL_IN_ONE_INSTALLER)
    set(FLANN_ROOT "${ROS_PCL_ROOT}/3rdParty/Flann")
  elseif(NOT FLANN_ROOT)
    get_filename_component(FLANN_ROOT "/usr/include" PATH)
  endif(ROS_PCL_ALL_IN_ONE_INSTALLER)
  if(PKG_CONFIG_FOUND)
    pkg_check_modules(PC_FLANN flann)
  endif(PKG_CONFIG_FOUND)

  find_path(FLANN_INCLUDE_DIRS flann/flann.hpp
    HINTS ${PC_FLANN_INCLUDEDIR} ${PC_FLANN_INCLUDE_DIRS} 
          "${FLANN_ROOT}" "$ENV{FLANN_ROOT}"
    PATHS "$ENV{PROGRAMFILES}/flann 1.6.9" "$ENV{PROGRAMW6432}/flann 1.6.9"
          "$ENV{PROGRAMFILES}/flann" "$ENV{PROGRAMW6432}/flann"
    PATH_SUFFIXES include)

  find_library(FLANN_LIBRARY
    NAMES flann_cpp_s flann_cpp
    HINTS ${PC_FLANN_LIBDIR} ${PC_FLANN_LIBRARY_DIRS} "${FLANN_ROOT}" "$ENV{FLANN_ROOT}"
    PATHS "$ENV{PROGRAMFILES}/flann 1.6.9" "$ENV{PROGRAMW6432}/flann 1.6.9" 
          "$ENV{PROGRAMFILES}/flann" "$ENV{PROGRAMW6432}/flann"
    PATH_SUFFIXES lib)

  find_library(FLANN_LIBRARY_DEBUG 
    NAMES flann_cpp_s-gd flann_cpp-gd flann_cpp_s flann_cpp
    HINTS ${PC_FLANN_LIBDIR} ${PC_FLANN_LIBRARY_DIRS} "${FLANN_ROOT}" "$ENV{FLANN_ROOT}"
    PATHS "$ENV{PROGRAMFILES}/flann 1.6.9" "$ENV{PROGRAMW6432}/flann 1.6.9" 
          "$ENV{PROGRAMFILES}/flann" "$ENV{PROGRAMW6432}/flann"
    PATH_SUFFIXES lib)

  find_package_handle_standard_args(Flann DEFAULT_MSG FLANN_LIBRARY FLANN_INCLUDE_DIRS)
  if(FLANN_FOUND)
    get_filename_component(FLANN_LIBRARY_PATH ${FLANN_LIBRARY} PATH)
    if(FLANN_LIBRARY_DEBUG)
      get_filename_component(FLANN_LIBRARY_DEBUG_PATH ${FLANN_LIBRARY_DEBUG} PATH)
      set(FLANN_LIBRARY_DIRS ${FLANN_LIBRARY_PATH} ${FLANN_LIBRARY_DEBUG_PATH}) 
      set(FLANN_LIBRARIES optimized ${FLANN_LIBRARY} debug ${FLANN_LIBRARY_DEBUG})
    else(FLANN_LIBRARY_DEBUG)
      set(FLANN_LIBRARY_DIRS ${FLANN_LIBRARY_PATH}) 
      set(FLANN_LIBRARIES ${FLANN_LIBRARY})
    endif(FLANN_LIBRARY_DEBUG)
    if("${FLANN_LIBRARY}" MATCHES "flann_cpp_s")
      set(FLANN_DEFINITIONS ${FLANN_DEFINITIONS} -DFLANN_STATIC)
    endif("${FLANN_LIBRARY}" MATCHES "flann_cpp_s")
  endif(FLANN_FOUND)
endmacro(find_flann)

macro(find_VTK)
  if(ROS_PCL_ALL_IN_ONE_INSTALLER)
    set(VTK_DIR "${ROS_PCL_ROOT}/3rdParty/VTK/lib/vtk-5.8")
  elseif(NOT VTK_DIR)
    set(VTK_DIR "/usr/lib64/vtk")
  endif(ROS_PCL_ALL_IN_ONE_INSTALLER)
  find_package(VTK ${QUIET_})
  if (VTK_FOUND AND NOT ANDROID_NDK)
    set(VTK_LIBRARIES vtkCommon vtkRendering vtkHybrid)
  endif(VTK_FOUND AND NOT ANDROID_NDK)
endmacro(find_VTK)

macro(find_libusb)
  if(NOT WIN32)
    find_path(LIBUSB_1_INCLUDE_DIR
      NAMES libusb-1.0/libusb.h
      PATHS /usr/include /usr/local/include /opt/local/include /sw/include
      PATH_SUFFIXES libusb-1.0)

    find_library(LIBUSB_1_LIBRARY
      NAMES usb-1.0
      PATHS /usr/lib /usr/local/lib /opt/local/lib /sw/lib)
    find_package_handle_standard_args(libusb-1.0 LIBUSB_1_LIBRARY LIBUSB_1_INCLUDE_DIR)
  endif(NOT WIN32)
endmacro(find_libusb)

# Finds each component external libraries if any
# The functioning is as following
# try to find _lib
# |--> _lib found ==> include the headers,
# |                   link to its library directories or include _lib_USE_FILE
# `--> _lib not found
#                   |--> _lib is optional ==> disable it (thanks to the guardians) 
#                   |                         and warn
#                   `--> _lib is required
#                                       |--> component is required explicitely ==> error
#                                       `--> component is induced ==> warn and remove it
#                                                                     from the list

macro(find_external_library _component _lib _is_optional)
  if("${_lib}" STREQUAL "boost")
    find_boost()
  elseif("${_lib}" STREQUAL "eigen")
    find_eigen()
  elseif("${_lib}" STREQUAL "flann")
    find_flann()
  elseif("${_lib}" STREQUAL "qhull")
    find_qhull()
  elseif("${_lib}" STREQUAL "openni")
    find_openni()
  elseif("${_lib}" STREQUAL "vtk")
    find_VTK()
  elseif("${_lib}" STREQUAL "libusb-1.0")
    find_libusb()
  endif("${_lib}" STREQUAL "boost")

  string(TOUPPER "${_component}" COMPONENT)
  string(TOUPPER "${_lib}" LIB)
  if(${LIB}_FOUND)
    list(APPEND ROS_PCL_${COMPONENT}_INCLUDE_DIRS ${${LIB}_INCLUDE_DIRS})
    if(${LIB}_USE_FILE)
      include(${${LIB}_USE_FILE})
    else(${LIB}_USE_FILE)
      list(APPEND ROS_PCL_${COMPONENT}_LIBRARY_DIRS "${${LIB}_LIBRARY_DIRS}")
    endif(${LIB}_USE_FILE)
    if(${LIB}_LIBRARIES)
      list(APPEND ROS_PCL_${COMPONENT}_LIBRARIES "${${LIB}_LIBRARIES}")
    endif(${LIB}_LIBRARIES)
    if(${LIB}_DEFINITIONS)
      list(APPEND ROS_PCL_${COMPONENT}_DEFINITIONS ${${LIB}_DEFINITIONS})
    endif(${LIB}_DEFINITIONS)
  else(${LIB}_FOUND)
    if("${_is_optional}" STREQUAL "OPTIONAL")
      add_definitions("-DDISABLE_${LIB}")
      ros_pcl_message("** WARNING ** ${_component} features related to ${_lib} will be disabled")
    elseif("${_is_optional}" STREQUAL "REQUIRED")
      if((NOT ROS_PCL_FIND_ALL) OR (ROS_PCL_FIND_ALL EQUAL 1))
        ros_pcl_report_not_found("${_component} is required but ${_lib} was not found")
      elseif(ROS_PCL_FIND_ALL EQUAL 0)
        # raise error and remove _component from ROS_PCL_TO_FIND_COMPONENTS
        string(TOUPPER "${_component}" COMPONENT)
        ros_pcl_message("** WARNING ** ${_component} will be disabled cause ${_lib} was not found")
        list(REMOVE_ITEM ROS_PCL_TO_FIND_COMPONENTS ${_component})
      endif((NOT ROS_PCL_FIND_ALL) OR (ROS_PCL_FIND_ALL EQUAL 1))
    endif("${_is_optional}" STREQUAL "OPTIONAL")
  endif(${LIB}_FOUND)
endmacro(find_external_library)

macro(ros_pcl_check_external_dependency _component)
endmacro(ros_pcl_check_external_dependency)

#flatten dependencies recursivity is great \o/
macro(compute_dependencies TO_FIND_COMPONENTS)
  foreach(component ${${TO_FIND_COMPONENTS}})
    set(ros_pcl_component ros_pcl_${component})
    if(${ros_pcl_component}_int_dep AND (NOT ROS_PCL_FIND_ALL))
      foreach(dependency ${${ros_pcl_component}_int_dep})
        list(FIND ${TO_FIND_COMPONENTS} ${component} pos)
        list(FIND ${TO_FIND_COMPONENTS} ${dependency} found)
        if(found EQUAL -1)
          set(ros_pcl_dependency ros_pcl_${dependency})
          if(${ros_pcl_dependency}_int_dep)
            list(INSERT ${TO_FIND_COMPONENTS} ${pos} ${dependency})
            if(ros_pcl_${dependency}_ext_dep)
              list(APPEND ros_pcl_${component}_ext_dep ${ros_pcl_${dependency}_ext_dep})
            endif(ros_pcl_${dependency}_ext_dep)
            if(ros_pcl_${dependency}_opt_dep)
              list(APPEND ros_pcl_${component}_opt_dep ${ros_pcl_${dependency}_opt_dep})
            endif(ros_pcl_${dependency}_opt_dep)
            compute_dependencies(${TO_FIND_COMPONENTS})
          else(${ros_pcl_dependency}_int_dep)
            list(INSERT ${TO_FIND_COMPONENTS} 0 ${dependency})
          endif(${ros_pcl_dependency}_int_dep)
        endif(found EQUAL -1)
      endforeach(dependency)
    endif(${ros_pcl_component}_int_dep AND (NOT ROS_PCL_FIND_ALL))
  endforeach(component)
endmacro(compute_dependencies)

### ---[ Find PCL

if(ROS_PCL_FIND_QUIETLY)
  set(QUIET_ QUIET)
else(ROS_PCL_FIND_QUIETLY)
  set(QUIET_)
endif(ROS_PCL_FIND_QUIETLY)

find_package(PkgConfig)

file(TO_CMAKE_PATH "${ROS_PCL_DIR}" ROS_PCL_DIR)
if(WIN32)
# ROS_PCLConfig.cmake is installed to ROS_PCL_ROOT/cmake
  get_filename_component(ROS_PCL_ROOT "${ROS_PCL_DIR}" PATH)
else(WIN32)
# ROS_PCLConfig.cmake is installed to ROS_PCL_ROOT/share/ros-pcl-x.y
  get_filename_component(ROS_PCL_ROOT "${ROS_PCL_DIR}" PATH)
  get_filename_component(ROS_PCL_ROOT "${ROS_PCL_ROOT}" PATH)
endif(WIN32)

# check whether ROS_PCLConfig.cmake is found into a PCL installation or in a build tree
if(EXISTS "${ROS_PCL_ROOT}/include/ros-pcl-${ROS_PCL_VERSION_MAJOR}.${ROS_PCL_VERSION_MINOR}/pcl/pcl_config.h")
  # Found a PCL installation
  # ros_pcl_message("Found a PCL installation")
  set(ROS_PCL_INCLUDE_DIRS "${ROS_PCL_ROOT}/include/ros-pcl-${ROS_PCL_VERSION_MAJOR}.${ROS_PCL_VERSION_MINOR}")
  set(ROS_PCL_LIBRARY_DIRS "${ROS_PCL_ROOT}/lib64")
  if(EXISTS "${ROS_PCL_ROOT}/3rdParty")
    set(ROS_PCL_ALL_IN_ONE_INSTALLER ON)
  endif(EXISTS "${ROS_PCL_ROOT}/3rdParty")
elseif(EXISTS "${ROS_PCL_DIR}/include/ros-pcl/pcl_config.h")
  # Found ROS_PCLConfig.cmake in a build tree of PCL
  # ros_pcl_message("PCL found into a build tree.")
  set(ROS_PCL_INCLUDE_DIRS "${ROS_PCL_DIR}/include") # for pcl_config.h
  set(ROS_PCL_LIBRARY_DIRS "${ROS_PCL_DIR}/lib64")
  set(ROS_PCL_SOURCES_TREE "/home/spot/rpmbuild/BUILD/ros-fuerte-ros-pcl-1.5.2/ros-pcl")
else(EXISTS "${ROS_PCL_ROOT}/include/ros-pcl-${ROS_PCL_VERSION_MAJOR}.${ROS_PCL_VERSION_MINOR}/pcl/pcl_config.h")
  ros_pcl_report_not_found("PCL can not be found on this machine")  
endif(EXISTS "${ROS_PCL_ROOT}/include/ros-pcl-${ROS_PCL_VERSION_MAJOR}.${ROS_PCL_VERSION_MINOR}/pcl/pcl_config.h")

#set a suffix for debug libraries
set(ROS_PCL_DEBUG_SUFFIX "")
set(ROS_PCL_RELEASE_SUFFIX "")

set(ros_pcl_all_components  common sample_consensus octree io kdtree search features filters keypoints surface registration segmentation tracking visualization )
list(LENGTH ros_pcl_all_components ROS_PCL_NB_COMPONENTS)

#list each component dependencies IN PCL
set(ros_pcl_sample_consensus_int_dep common )
set(ros_pcl_octree_int_dep common )
set(ros_pcl_io_int_dep common octree )
set(ros_pcl_kdtree_int_dep common )
set(ros_pcl_search_int_dep common kdtree octree )
set(ros_pcl_features_int_dep common search kdtree octree )
set(ros_pcl_filters_int_dep common sample_consensus search kdtree octree )
set(ros_pcl_keypoints_int_dep common search kdtree octree features filters )
set(ros_pcl_surface_int_dep common search kdtree octree )
set(ros_pcl_registration_int_dep common kdtree search sample_consensus features )
set(ros_pcl_segmentation_int_dep common search sample_consensus kdtree octree )
set(ros_pcl_tracking_int_dep common search kdtree filters octree )
set(ros_pcl_visualization_int_dep common io kdtree )


#list each component external dependencies (ext means mandatory and opt means optional)
set(ros_pcl_common_ext_dep eigen boost )
set(ros_pcl_kdtree_ext_dep flann )
set(ros_pcl_search_ext_dep flann )
set(ros_pcl_visualization_ext_dep vtk )


set(ros_pcl_io_opt_dep openni vtk libusb-1.0 )
set(ros_pcl_surface_opt_dep qhull )
set(ros_pcl_visualization_opt_dep openni )


set(ros_pcl_header_only_components cuda_common)

include(FindPackageHandleStandardArgs)

#check if user provided a list of components
#if no components at all or full list is given set ROS_PCL_FIND_ALL
if(ROS_PCL_FIND_COMPONENTS)
  list(LENGTH ROS_PCL_FIND_COMPONENTS ROS_PCL_FIND_COMPONENTS_LENGTH)
  if(ROS_PCL_FIND_COMPONENTS_LENGTH EQUAL ROS_PCL_NB_COMPONENTS)
    set(ROS_PCL_TO_FIND_COMPONENTS ${ros_pcl_all_components})
    set(ROS_PCL_FIND_ALL 1)
  else(ROS_PCL_FIND_COMPONENTS_LENGTH EQUAL ROS_PCL_NB_COMPONENTS)
    set(ROS_PCL_TO_FIND_COMPONENTS ${ROS_PCL_FIND_COMPONENTS})    
  endif(ROS_PCL_FIND_COMPONENTS_LENGTH EQUAL ROS_PCL_NB_COMPONENTS)
else(ROS_PCL_FIND_COMPONENTS)
  set(ROS_PCL_TO_FIND_COMPONENTS ${ros_pcl_all_components})
  set(ROS_PCL_FIND_ALL 1)
endif(ROS_PCL_FIND_COMPONENTS)

compute_dependencies(ROS_PCL_TO_FIND_COMPONENTS)

# compute external dependencies per component
foreach(component ${ROS_PCL_TO_FIND_COMPONENTS})
    foreach(opt ${ros_pcl_${component}_opt_dep})
      find_external_library(${component} ${opt} OPTIONAL)
    endforeach(opt)
    foreach(ext ${ros_pcl_${component}_ext_dep})
      find_external_library(${component} ${ext} REQUIRED)
    endforeach(ext) 
endforeach(component)

foreach(component ${ROS_PCL_TO_FIND_COMPONENTS})
  set(ros_pcl_component ros_pcl_${component})
  string(TOUPPER "${component}" COMPONENT)

  ros_pcl_message(STATUS "looking for ROS_PCL_${COMPONENT}")

  string(REGEX REPLACE "^cuda_(.*)$" "\\1" cuda_component "${component}")
  string(REGEX REPLACE "^gpu_(.*)$" "\\1" gpu_component "${component}") 
  
  find_path(ROS_PCL_${COMPONENT}_INCLUDE_DIR
    NAMES pcl/${component}
          pcl/cuda/${cuda_component} pcl/cuda/${gpu_component} pcl/cuda/${component}
          pcl/gpu/${cuda_component} pcl/gpu/${gpu_component} pcl/gpu/${component}
    HINTS ${ROS_PCL_INCLUDE_DIRS}
          "${ROS_PCL_SOURCES_TREE}"
    PATH_SUFFIXES
          ${component}/include
          cuda/${cuda_component}/include 
          gpu/${gpu_component}/include
    DOC "path to ${component} headers"
    NO_DEFAULT_PATH)

  if(ROS_PCL_${COMPONENT}_INCLUDE_DIR)
    list(APPEND ROS_PCL_${COMPONENT}_INCLUDE_DIRS "${ROS_PCL_${COMPONENT}_INCLUDE_DIR}")
  else(ROS_PCL_${COMPONENT}_INCLUDE_DIR)
    #ros_pcl_message("No include directory found for ros_pcl_${component}.")
  endif(ROS_PCL_${COMPONENT}_INCLUDE_DIR)
  
  # Skip find_library for header only modules
  list(FIND ros_pcl_header_only_components ${component} _is_header_only)
  if(_is_header_only EQUAL -1)
    find_library(ROS_PCL_${COMPONENT}_LIBRARY ${ros_pcl_component}${ROS_PCL_RELEASE_SUFFIX}
      HINTS ${ROS_PCL_LIBRARY_DIRS}
      DOC "path to ${ros_pcl_component} library"
      NO_DEFAULT_PATH)
    get_filename_component(${component}_library_path 
      ${ROS_PCL_${COMPONENT}_LIBRARY}
      PATH)

    find_library(ROS_PCL_${COMPONENT}_LIBRARY_DEBUG ${ros_pcl_component}${ROS_PCL_DEBUG_SUFFIX}
      HINTS ${ROS_PCL_LIBRARY_DIRS}
      DOC "path to ${ros_pcl_component} library debug"
      NO_DEFAULT_PATH)
    if(ROS_PCL_${COMPONENT}_LIBRARY_DEBUG)
      get_filename_component(${component}_library_path_debug 
        ${ROS_PCL_${COMPONENT}_LIBRARY_DEBUG}
        PATH)
    endif(ROS_PCL_${COMPONENT}_LIBRARY_DEBUG)

    find_package_handle_standard_args(ROS_PCL_${COMPONENT} DEFAULT_MSG
      ROS_PCL_${COMPONENT}_LIBRARY ROS_PCL_${COMPONENT}_INCLUDE_DIR)
  else(_is_header_only EQUAL -1)
    find_package_handle_standard_args(ROS_PCL_${COMPONENT} DEFAULT_MSG
      ROS_PCL_${COMPONENT}_INCLUDE_DIR)  
  endif(_is_header_only EQUAL -1)
  
  if(ROS_PCL_${COMPONENT}_FOUND)
    if(NOT "${ROS_PCL_${COMPONENT}_INCLUDE_DIRS}" STREQUAL "")
      list(REMOVE_DUPLICATES ROS_PCL_${COMPONENT}_INCLUDE_DIRS)
    endif(NOT "${ROS_PCL_${COMPONENT}_INCLUDE_DIRS}" STREQUAL "")
    list(APPEND ROS_PCL_INCLUDE_DIRS ${ROS_PCL_${COMPONENT}_INCLUDE_DIRS})
    mark_as_advanced(ROS_PCL_${COMPONENT}_INCLUDE_DIRS)
    if(_is_header_only EQUAL -1)
      list(APPEND ROS_PCL_DEFINITIONS ${ROS_PCL_${COMPONENT}_DEFINITIONS})
      list(APPEND ROS_PCL_LIBRARY_DIRS ${component_library_path})
      if(ROS_PCL_${COMPONENT}_LIBRARY_DEBUG)
        list(APPEND ROS_PCL_${COMPONENT}_LIBRARIES optimized ${ROS_PCL_${COMPONENT}_LIBRARY} debug ${ROS_PCL_${COMPONENT}_LIBRARY_DEBUG})
        list(APPEND ROS_PCL_LIBRARY_DIRS ${component_library_path_debug})
      else(ROS_PCL_${COMPONENT}_LIBRARY_DEBUG)
        list(APPEND ROS_PCL_${COMPONENT}_LIBRARIES ${ROS_PCL_${COMPONENT}_LIBRARY})
      endif(ROS_PCL_${COMPONENT}_LIBRARY_DEBUG)
      list(APPEND ROS_PCL_LIBRARIES ${ROS_PCL_${COMPONENT}_LIBRARIES})
      mark_as_advanced(ROS_PCL_${COMPONENT}_LIBRARY ROS_PCL_${COMPONENT}_LIBRARY_DEBUG)
    endif(_is_header_only EQUAL -1)    
    # Append internal dependencies
    foreach(int_dep ${ros_pcl_${component}_int_dep})
      string(TOUPPER "${int_dep}" INT_DEP)
      if(ROS_PCL_${INT_DEP}_FOUND)
        list(APPEND ROS_PCL_${COMPONENT}_INCLUDE_DIRS ${ROS_PCL_${INT_DEP}_INCLUDE_DIRS})
        if(ROS_PCL_${INT_DEP}_LIBRARIES)
          list(APPEND ROS_PCL_${COMPONENT}_LIBRARIES "${ROS_PCL_${INT_DEP}_LIBRARIES}")
        endif(ROS_PCL_${INT_DEP}_LIBRARIES)	
      endif(ROS_PCL_${INT_DEP}_FOUND)
    endforeach(int_dep)
  endif(ROS_PCL_${COMPONENT}_FOUND)
endforeach(component)

if(NOT "${ROS_PCL_INCLUDE_DIRS}" STREQUAL "")
  list(REMOVE_DUPLICATES ROS_PCL_INCLUDE_DIRS)
endif(NOT "${ROS_PCL_INCLUDE_DIRS}" STREQUAL "")

if(NOT "${ROS_PCL_LIBRARY_DIRS}" STREQUAL "")
  list(REMOVE_DUPLICATES ROS_PCL_LIBRARY_DIRS)
endif(NOT "${ROS_PCL_LIBRARY_DIRS}" STREQUAL "")

# We need to export march=native for tutorials or user code
list (APPEND ROS_PCL_DEFINITIONS " -march=native")

if(NOT "${ROS_PCL_DEFINITIONS}" STREQUAL "")
  list(REMOVE_DUPLICATES ROS_PCL_DEFINITIONS)
endif(NOT "${ROS_PCL_DEFINITIONS}" STREQUAL "")

ros_pcl_remove_duplicate_libraries(ROS_PCL_LIBRARIES ROS_PCL_DEDUP_LIBRARIES)
set(ROS_PCL_LIBRARIES ${ROS_PCL_DEDUP_LIBRARIES})

find_package_handle_standard_args(PCL DEFAULT_MSG ROS_PCL_LIBRARIES ROS_PCL_INCLUDE_DIRS)
mark_as_advanced(ROS_PCL_LIBRARIES ROS_PCL_INCLUDE_DIRS ROS_PCL_LIBRARY_DIRS)
